#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 10.08.2017 20:33
:Licence GNUv3
Part of grammpy-transforms

"""

from typing import List
from grammpy import *
from .NongeneratingSymbolsRemove import *
from .UnreachableSymbolsRemove import *
from .EpsilonRulesRemove import *
from .UnitRulesRemove import *
from .ChomskyForm import *

__all__ = ['ContextFree']


class ContextFree:

    EpsilonRemovedRule = EpsilonRemovedRule
    ReducedUnitRule = ReducedUnitRule
    ChomskyRule = ChomskyRule
    ChomskySplitRule = ChomskySplitRule
    ChomskyTermRule = ChomskyTermRule
    ChomskyRestRule = ChomskyRestRule
    ChomskyTerminalReplaceRule = ChomskyTerminalReplaceRule

    ChomskyNonterminal = ChomskyNonterminal
    ChomskyGroupNonterminal = ChomskyGroupNonterminal
    ChomskyTermNonterminal = ChomskyTermNonterminal

    UnitSymbolRechablingResults = UnitSymbolRechablingResults

    @staticmethod
    def remove_nongenerating_nonterminals(grammar: Grammar, transform_grammar=False):
        return remove_nongenerating_nonterminals(grammar, transform_grammar)

    @staticmethod
    def is_grammar_generating(grammar: Grammar, transform_grammar=False, perform_remove=True):
        g = grammar
        if perform_remove:
            g = ContextFree.remove_nongenerating_nonterminals(grammar, transform_grammar=transform_grammar)
        return g.start_get() in g.nonterms()

    @staticmethod
    def remove_unreachable_symbols(grammar: Grammar, transform_grammar=False):
        return remove_unreachable_symbols(grammar, transform_grammar)

    @staticmethod
    def remove_useless_symbols(grammar: Grammar, transform_grammar=False, *,
                                   perform_unreachable_alg = True,
                                   perform_nongenerating_alg = True) -> Grammar:
        if perform_nongenerating_alg:
            grammar = ContextFree.remove_nongenerating_nonterminals(grammar, transform_grammar=transform_grammar)
            transform_grammar = True
        if perform_unreachable_alg:
            grammar = ContextFree.remove_unreachable_symbols(grammar, transform_grammar=transform_grammar)
        return grammar

    @staticmethod
    def remove_rules_with_epsilon(grammar: Grammar, transform_grammar=False) -> Grammar:
        return remove_rules_with_epsilon(grammar, transform_grammar=transform_grammar)

    @staticmethod
    def find_nonterminals_rewritable_to_epsilon(grammar: Grammar) -> List[Nonterminal]:
        return find_nonterminals_rewritable_to_epsilon(grammar)

    @staticmethod
    def find_nonterminals_reachable_by_unit_rules(grammar: Grammar) -> UnitSymbolRechablingResults:
        return find_nonterminals_reachable_by_unit_rules(grammar)

    @staticmethod
    def remove_unit_rules(grammar: Grammar, transform_grammar=False) -> Grammar:
        return remove_unit_rules(grammar, transform_grammar=transform_grammar)

    @staticmethod
    def transform_to_chomsky_normal_form(grammar: Grammar, transform_grammar=False) -> Grammar:
        return transform_to_chomsky_normal_form(grammar, transform_grammar=transform_grammar)


#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 22.08.2017 11:39
:Licence GNUv3
Part of grammpy-transforms

"""

from inspect import isclass
from typing import List, Dict, Optional
from grammpy import Grammar, Nonterminal, Rule


class UnitSymbolRechablingResults:
    def __init__(self, field, translation):
        self.f = field  # type: List[List[List[Rule] | None]]
        self.t = translation  # type: Dict[Nonterminal, int]

    def reach(self, from_symbol: Nonterminal, to_symbol: Nonterminal) -> bool:
        if from_symbol not in self.t or to_symbol not in self.t:
            return False
        return self.f[self.t[from_symbol]][self.t[to_symbol]] is not None

    def reachables(self, from_symbol: Nonterminal) -> List[Nonterminal]:
        if from_symbol not in self.t:
            return []
        reachable = []
        index = self.t[from_symbol]
        for n, i in self.t.items():
            if self.f[index][i] is not None:
                reachable.append(n)
        return reachable

    def path_rules(self, from_symbol: Nonterminal, to_symbol: Nonterminal):
        if from_symbol not in self.t or to_symbol not in self.t:
            return None
        return self.f[self.t[from_symbol]][self.t[to_symbol]]


def find_nonterminals_reachable_by_unit_rules(grammar: Grammar) -> UnitSymbolRechablingResults:
    # Get nonterms
    nonterms = grammar.nonterms()
    l = len(nonterms)
    # Create indexes for nonterms
    nonterm_to_index = dict()
    for i in range(l):
        nonterm_to_index[nonterms[i]] = i
    # Prepare field
    field = [[None for _ in nonterms] for _ in nonterms]
    # Fill diagonal
    for i in range(l):
        field[i][i] = list()
    # Fill rules
    for rule in grammar.rules():
        if len(rule.left)==1 and len(rule.right)==1 and \
                isclass(rule.fromSymbol) and isclass(rule.toSymbol) and \
                issubclass(rule.fromSymbol, Nonterminal) and issubclass(rule.toSymbol, Nonterminal):
            field[nonterm_to_index[rule.fromSymbol]][nonterm_to_index[rule.toSymbol]] = [rule]
    # Floyd Warshall
    f = field
    for k in range(l):
        for i in range(l):
            for j in range(l):
                if f[i][k] is not None and f[k][j] is not None:
                    if f[i][j] is None or len(f[i][j]) > len(f[i][k]) + len(f[k][j]):
                        f[i][j] = f[i][k] + f[k][j]
    # Return results
    return UnitSymbolRechablingResults(f, nonterm_to_index)





#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 17.08.207 13:29
:Licence GNUv3
Part of grammpy-transforms

"""

from grammpy import Grammar
from grammpy.exceptions import NotNonterminalException
from copy import copy

class StartSymbolNotSpecifiedException(Exception):
    pass


def remove_unreachable_symbols(grammar: Grammar, transform_grammar=False) -> Grammar:
    # Copy if required
    if transform_grammar is False: grammar = copy(grammar)
    # Check if start symbol is set
    if not grammar.start_isSet(): raise StartSymbolNotSpecifiedException()
    # Create process sets
    reachable = {grammar.start_get()}
    rules = grammar.rules()
    # Begin iterations
    while True:
        # Create sets for current iteration
        active = reachable.copy()
        processedRules = []
        # Loop rest of rules
        for rule in rules:
            # If left part of rule already in reachable symbols
            if rule.fromSymbol in reachable:
                # Set symbols as reachable
                processedRules.append(rule)
                for symbol in rule.right: active.add(symbol)
        # End of rules loop
        # Remove processed rules
        for item in processedRules: rules.remove(item)
        # If current and previous iterations are same, than end iterations
        if active == reachable: break
        reachable = active
    # End of iterations
    # Set symbols to remove
    allSymbols = set(grammar.nonterms()).union(set(x.s for x in grammar.terms()))
    for symbol in allSymbols.difference(reachable):
        try:
            grammar.remove_nonterm(symbol)
        except NotNonterminalException:
            grammar.remove_term(symbol)
    return grammar



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 10.08.2017 20:32
:Licence GNUv3
Part of grammpy-transforms

"""

from .transform_to_chomsky_normal_form import *
from .transform_from_chomsky_normal_form import transform_from_chomsky_normal_form


#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 10.08.2017 20:32
:Licence GNUv3
Part of grammpy-transforms

"""

from .remove_nongenerating_nonterminals import remove_nongenerating_nonterminals


#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 10.08.2017 20:33
:Licence GNUv3
Part of grammpy-transforms

"""

from copy import copy
from grammpy import Grammar, EPSILON

def remove_nongenerating_nonterminals(grammar: Grammar, transform_grammar=False) -> Grammar:
    # Copy if required
    if transform_grammar is False: grammar = copy(grammar)
    # Create working sets
    generates = set(item.s for item in grammar.terms())
    generates.add(EPSILON)
    rules = set(rule for rule in grammar.rules())
    while True:
        # Create set of next iteration
        additional = generates.copy()
        processedRules = []
        # Iterate over unprocessed rules
        for rule in rules:
            rightPart = rule.right
            allIn = True
            # Check if all symbols on the right part of rule are in generates set
            for symbol in rightPart:
                if symbol not in generates:
                    allIn = False
                    break
            # Symbol is missing so rule is not process
            if not allIn: continue
            # Rule is process - remove it from processing rules and make symbol as generating
            additional.add(rule.fromSymbol)
            processedRules.append(rule)
        # End of rules iterations
        # Remove process rules in current iteration
        for item in processedRules: rules.remove(item)
        # If current and previous iterations are same, than end iterations
        if additional == generates: break
        # Swap sets from previous and current iterations
        generates = additional
    # Remove nonterms that are not generating
    for nonterm in set(grammar.nonterms()).difference(generates):
        grammar.remove_nonterm(nonterm)
    return grammar



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 03.09.2017 10:49
:Licence GNUv3
Part of grammpy-transforms

"""

from grammpy import *
from .UnitRulesRemove import unit_rules_restore
from .EpsilonRulesRemove import epsilon_rules_restore
from .ChomskyForm import transform_from_chomsky_normal_form

__all__ = ['InverseContextFree']


class InverseContextFree:
    @staticmethod
    def unit_rules_restore(root: Nonterminal) -> Nonterminal:
        return unit_rules_restore(root)

    @staticmethod
    def epsilon_rules_restore(root: Nonterminal) -> Nonterminal:
        return epsilon_rules_restore(root)

    @staticmethod
    def transform_from_chomsky_normal_form(root: Nonterminal) -> Nonterminal:
        return transform_from_chomsky_normal_form(root)



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 22.08.2017 11:37
:Licence GNUv3
Part of grammpy-transforms

"""

from inspect import isclass
from copy import copy
from typing import List
from grammpy import *
from .find_symbols_reachable_by_unit_rules import find_nonterminals_reachable_by_unit_rules


class ReducedUnitRule(Rule):
    by_rules = []  # type: List[type]
    end_rule = None  # type: Rule


def _is_unit(rule):
    return len(rule.left) == 1 and len(rule.right) == 1 and \
           isclass(rule.fromSymbol) and isclass(rule.toSymbol) and \
           issubclass(rule.fromSymbol, Nonterminal) and issubclass(rule.toSymbol, Nonterminal)


def _create_rule(path, rule):
    created = type("Reduced" + rule.__name__, (ReducedUnitRule,), ReducedUnitRule.__dict__.copy())
    created.rule = ([path[0].fromSymbol], rule.right)
    created.end_rule = rule
    created.by_rules = path
    return created


def remove_unit_rules(grammar, transform_grammar=False):
    if transform_grammar is False: grammar = copy(grammar)
    # get connections
    res = find_nonterminals_reachable_by_unit_rules(grammar)
    # iterate through rules
    for rule in grammar.rules():
        if _is_unit(rule):
            grammar.remove_rule(rule)
            continue
        for nonterm in grammar.nonterms():
            path = res.path_rules(nonterm, rule.fromSymbol)
            if path is not None and len(path) > 0:
                created = _create_rule(path, rule)
                grammar.add_rule(created)
    return grammar



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 03.09.2017 10:46
:Licence GNUv3
Part of grammpy-transforms

"""
from collections import deque

from grammpy import *
from .transform_to_chomsky_normal_form import *
from ..Manipulations import Traversing, Manipulations

def transform_from_chomsky_normal_form(root: Nonterminal):
    items = Traversing.postOrder(root)
    items = filter(lambda x: isinstance(x, (ChomskyTermRule,ChomskyTerminalReplaceRule, ChomskySplitRule)), items)
    de = deque(items)
    while de:
        rule = de.popleft()
        if isinstance(rule, ChomskyTermRule):
            upper_nonterm = rule.from_symbols[0]  # type: Nonterminal
            to_append = upper_nonterm.from_rule  # type: Rule
            index = to_append.to_symbols.index(upper_nonterm)
            term = rule.to_symbols[0]  # type: Terminal
            to_append.to_symbols[index] = term
            term._set_from_rule(to_append)
        elif isinstance(rule, ChomskyTerminalReplaceRule):
            to_rule = rule.from_rule()  # type: Rule
            for p in rule.from_symbols:  # type: Nonterminal
                to_rule._from_symbols.append(p)
                p._set_to_rule(to_rule)
            for c in rule.to_symbols:  # type: Nonterminal
                to_rule._to_symbols.append(c)
                c._set_from_rule(to_rule)
            de.append(to_rule)
        elif isinstance(rule, ChomskySplitRule):
            created_rule = rule.from_rule()  # type: Rule
            #parent nonterminals
            for p in rule.from_symbols:  # type: Nonterminal
                p._set_to_rule(created_rule)
                created_rule._from_symbols.append(p)
            #left child
            left_child = rule.to_symbols[0] # type: Nonterminal
            left_child._set_from_rule(created_rule)
            created_rule._to_symbols.append(left_child)
            #right childs
            for ch in rule.to_symbols[1].to_rule.to_symbols:  # type: Nonterminal
                ch._set_from_rule(created_rule)
                created_rule.to_symbols.append(ch)
            de.append(created_rule)
    return root


#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 24.12.2017 11:31
:Licence GNUv3
Part of grammpy-transforms

"""
import functools
import operator
from typing import Callable

from grammpy import Rule, Nonterminal, Terminal


class Manipulations:
    @staticmethod
    def replaceRule(oldRule: Rule, newRule: Rule) -> Rule:
        for par in oldRule.from_symbols:
            par._set_to_rule(newRule)
            newRule._from_symbols.append(par)
        for ch in oldRule.to_symbols:
            ch._set_from_rule(newRule)
            newRule._to_symbols.append(ch)
        return newRule

    @staticmethod
    def replaceNode(oldNode: Nonterminal, newNode: Nonterminal):
        if oldNode.from_rule is not None and len(oldNode.from_rule.to_symbols) > 0:
            indexParent = oldNode.from_rule.to_symbols.index(oldNode)
            oldNode.from_rule.to_symbols[indexParent] = newNode
            newNode._set_from_rule(oldNode.from_rule)
        if oldNode.to_rule is not None and len(oldNode.to_rule.from_symbols) > 0:
            indexChild = oldNode.to_rule.from_symbols.index(oldNode)
            oldNode.to_rule._from_symbols[indexChild] = newNode
            newNode._set_to_rule(oldNode.to_rule)
        return newNode

    @staticmethod
    def replace(oldEl, newEl):
        if isinstance(oldEl, Rule):
            return Manipulations.replaceRule(oldEl, newEl)
        if isinstance(oldEl, (Nonterminal, Terminal)):
            return Manipulations.replaceNode(oldEl, newEl)

class Traversing:
    @staticmethod
    def traverse(root, callback: Callable):
        def innerCallback(item):
            return callback(item, innerCallback)
        return callback(root, innerCallback)

    @staticmethod
    def traverseSeparated(root, callbackRules, callbackNonterminals, callbackTerminals):
        def separateTraverse(item, callback):
            if isinstance(item, Rule):
                return callbackRules(item, callback)
            if isinstance(item, Nonterminal):
                return callbackNonterminals(item, callback)
            if isinstance(item, Terminal):
                return callbackTerminals(item, callback)
        return Traversing.traverse(root, separateTraverse)

    @staticmethod
    def preOrder(root):
        def travRule(item, callback):
            resp = [callback(ch) for ch in item.to_symbols]
            return functools.reduce(operator.add, resp, [item])
        def travNonterminals(item, callback):
            return [item] + callback(item.to_rule)
        def travTerms(item, callback):
            return [item]
        return Traversing.traverseSeparated(root, travRule, travNonterminals, travTerms)

    @staticmethod
    def postOrder(root):
        def travRule(item, callback):
            resp = [callback(ch) for ch in item.to_symbols]
            return functools.reduce(operator.add, resp, []) + [item]
        def travNonterminals(item, callback):
            return callback(item.to_rule) + [item]
        def travTerms(item, callback):
            return [item]
        return Traversing.traverseSeparated(root, travRule, travNonterminals, travTerms)



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 10.08.2017 20:32
:Licence GNUv3
Part of grammpy-transforms

"""

from .ContextFree import ContextFree
from .InverseContextFree import InverseContextFree
from .InverseCommon import InverseCommon
from .Manipulations import Manipulations, Traversing



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 28.08.2017 10:19
:Licence GNUv3
Part of grammpy-transforms

"""

from inspect import isclass
from copy import copy
from grammpy import *


class ChomskyNonterminal(Nonterminal): pass
class ChomskyGroupNonterminal(ChomskyNonterminal):
    group = []
class ChomskyTermNonterminal(ChomskyNonterminal):
    for_term = None
class ChomskyRule(Rule): pass
class ChomskySplitRule(ChomskyRule):
    from_rule = None
class ChomskyRestRule(ChomskyRule):
    from_rule = None
class ChomskyTerminalReplaceRule(ChomskyRule):
    from_rule = None
    replaced_index = None
class ChomskyTermRule(ChomskyRule): pass

class Container:
    def __init__(self, terminal, nonterminal, rule):
        self.used = False
        self.terminal = terminal
        self.nonterminal = nonterminal
        self.rule = rule

class TerminalsFilling:
    def __init__(self, grammar: Grammar):
        self._grammar = grammar
        self._items = dict()
        self._counter = 0
        for term in grammar.terms():
            t = term.s
            created_nonterm = type("NonterminalForTerm" + str(self._counter), (ChomskyTermNonterminal,), ChomskyTermNonterminal.__dict__.copy())
            created_nonterm.for_term = t
            created_rule = type("RuleForTerm" + str(self._counter), (ChomskyTermRule,), ChomskyTermRule.__dict__.copy())
            created_rule.rule = ([created_nonterm], [t])
            self._items[t] = Container(term, created_nonterm, created_rule)
            self._counter += 1

    def get(self, term):
        if self._items[term].used is False:
            cont = self._items[term]
            self._grammar.add_nonterm(cont.nonterminal)
            self._grammar.add_rule(cont.rule)
            cont.used = True
        return self._items[term].nonterminal


def transform_to_chomsky_normal_form(grammar: Grammar, transform_grammar=False):
    # Copy if required
    if transform_grammar is False: grammar = copy(grammar)
    fill = TerminalsFilling(grammar)
    all_rules = grammar.rules()
    index = 0
    while index < len(all_rules):
        rule = all_rules[index]  # type: Rule
        index += 1
        #Check, if rule must be split
        if len(rule.right) > 2:
            grammar.remove_rule(rule)
            #create nonterm that represent group on the right
            created_nonterm = type("GroupNonterminal"+str(index-1), (ChomskyGroupNonterminal,), ChomskyGroupNonterminal.__dict__.copy())
            created_nonterm.group = rule.right[1:]
            #create rule that replace current
            created_left_rule = type("SplitRule"+str(index-1), (ChomskySplitRule,), ChomskySplitRule.__dict__.copy())
            created_left_rule.rule = ([rule.fromSymbol], [rule.right[0], created_nonterm])
            created_left_rule.from_rule = rule
            #create rule with symbols on the right
            created_right_rule = type("SplitTempRule" + str(index - 1), (ChomskyRestRule,), ChomskyRestRule.__dict__.copy())
            created_right_rule.rule = ([created_nonterm], rule.right[1:])
            created_right_rule.from_rule = rule
            #fill
            grammar.add_nonterm(created_nonterm)
            grammar.add_rule([created_left_rule, created_right_rule])
            all_rules.append(created_left_rule)
            all_rules.append(created_right_rule)
        # Check, if must replace terminal
        elif len(rule.right) == 2:
            if grammar.have_term(rule.right[0]):
                #first symbol is terminal
                grammar.remove_rule(rule)
                symb = fill.get(rule.right[0])
                created = type("LeftReplaceRule"+str(index-1), (ChomskyTerminalReplaceRule,), ChomskyTerminalReplaceRule.__dict__.copy())
                created.rule = ([rule.fromSymbol], [symb, rule.right[1]])
                created.from_rule = rule
                created.replaced_index = 0
                # fill it
                grammar.add_rule(created)
                all_rules.append(created)
            elif grammar.have_term(rule.right[1]):
                #second symbol is terminal
                grammar.remove_rule(rule)
                symb = fill.get(rule.right[1])
                created = type("RightReplaceRule" + str(index - 1), (ChomskyTerminalReplaceRule,), ChomskyTerminalReplaceRule.__dict__.copy())
                created.rule = ([rule.fromSymbol], [rule.right[0], symb])
                created.from_rule = rule
                created.replaced_index = 1
                # fill it
                grammar.add_rule(created)
                all_rules.append(created)
    return grammar



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 20.08.2017 15:42
:Licence GNUv3
Part of grammpy-transforms

"""

from grammpy import Grammar, EPSILON


def find_nonterminals_rewritable_to_epsilon(grammar: Grammar) -> dict:
    rewritable = dict()
    rewritable[EPSILON] = None
    while True:
        working = rewritable.copy()
        for rule in grammar.rules():
            allRewritable = True
            for symbol in rule.right:
                if symbol not in rewritable: allRewritable = False
            if allRewritable and rule.fromSymbol not in working:
                working[rule.fromSymbol] = rule
        if working == rewritable: break
        rewritable = working
    del rewritable[EPSILON]
    return rewritable



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 03.09.2017 10:47
:Licence GNUv3
Part of grammpy-transforms

"""

from grammpy import *
from .remove_rules_with_epsilon import EpsilonRemovedRule
from ..Manipulations import Manipulations, Traversing


def _restore_tree_for(root: Nonterminal, translate: dict):
    if root is EPSILON:
        return EPSILON
    created_nonterm = root()  # type: Nonterminal
    created_rule = translate[root]()  # type: Rule
    created_nonterm._set_to_rule(created_rule)
    created_rule._from_symbols.append(created_nonterm)
    for ch in created_rule.right:
        p = _restore_tree_for(ch, translate)  # type: Nonterminal
        if p is not EPSILON:
            p._set_from_rule(created_rule)
        created_rule._to_symbols.append(p)
    return created_nonterm

def epsilon_rules_restore(root: Nonterminal):
    items = Traversing.postOrder(root)
    items = filter(lambda x: isinstance(x, EpsilonRemovedRule), items)
    for rule in items:
        created_rule = rule.from_rule()  # type: Rule
        #Solve parents
        for s in rule.from_symbols:  # type: Nonterminal
            s._set_to_rule(created_rule)
            created_rule._from_symbols.append(s)
        #Copy childs to replace index
        for i in range(rule.replace_index):
            ch = rule.to_symbols[i]  # type: Nonterminal
            ch._set_from_rule(created_rule)
            created_rule._to_symbols.append(ch)
        #Add epsilon
        symb = _restore_tree_for(created_rule.right[rule.replace_index], rule.backtrack) # type: Nonterminal
        created_rule._to_symbols.append(symb)
        if symb is not EPSILON:
            symb._set_from_rule(created_rule)
        #Add rest of childs
        for i in range(rule.replace_index, len(rule.to_symbols)):
            ch = rule.to_symbols[i]  # type: Nonterminal
            ch._set_from_rule(created_rule)
            created_rule._to_symbols.append(ch)
    return root



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 23.12.2017 16:03
:Licence GNUv3
Part of grammpy-transforms

"""

from .SplittedRules import *

class InverseCommon:

    @staticmethod
    def splitted_rules(root):
        return splitted_rules(root)



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 10.08.2017 20:32
:Licence GNUv3
Part of grammpy-transforms

"""

from .find_nonterminals_rewritable_to_epsilon import find_nonterminals_rewritable_to_epsilon
from .remove_rules_with_epsilon import EpsilonRemovedRule, remove_rules_with_epsilon
from .epsilon_rules_restore import epsilon_rules_restore



from setuptools import setup

v = '1.2.0'

setup(
    name='grammpy-transforms',
    version=v,
    packages=['grammpy_transforms',
              'grammpy_transforms.ChomskyForm', 'grammpy_transforms.UnitRulesRemove',
              'grammpy_transforms.EpsilonRulesRemove', 'grammpy_transforms.UnreachableSymbolsRemove',
              'grammpy_transforms.NongeneratingSymbolsRemove'],
    url='https://github.com/PatrikValkovic/grammpy-transforms',
    download_url='https://github.com/PatrikValkovic/grammpy-transforms/archive/v' + v + '.tar.gz',
    license='GNU General Public License v3.0',
    author='Patrik Valkovic',
    author_email='patrik.valkovic@hotmail.cz',
    description='Set of transformations for grammpy library.',
    install_requires=[
        'grammpy',
    ],
)



#!/usr/bin/env python
"""
:Author Patrik Valkovic
:Created 20.08.2017 15:43
:Licence GNUv3
Part of grammpy-transforms

"""

from copy import copy
from grammpy import Grammar, Rule, EPSILON
from .find_nonterminals_rewritable_to_epsilon import find_nonterminals_rewritable_to_epsilon


class EpsilonRemovedRule(Rule):
    from_rule = None  # type: Rule
    replace_index = None  # type: int
    backtrack = None  # type: dict

def _create_rule(rule: Rule, index: int, backtrack: dict) -> EpsilonRemovedRule:
    # Remove old rules
    old_dict = rule.__dict__.copy()
    if 'rules' in old_dict: del old_dict['rules']
    if 'rule' in old_dict: del old_dict['rule']
    if 'left' in old_dict: del old_dict['left']
    if 'right' in old_dict: del old_dict['right']
    if 'fromSymbol' in old_dict: del old_dict['fromSymbol']
    if 'toSymbol' in old_dict: del old_dict['toSymbol']
    # Create type
    created = type('NoEps'+rule.__name__, (EpsilonRemovedRule,), old_dict)
    # Add from_rule and index
    created.from_rule = rule
    created.replace_index = index
    created.backtrack = backtrack
    # Add rule
    created.fromSymbol = rule.fromSymbol
    created.right = [rule.right[i] for i in range(len(rule.right)) if i != index]
    if len(created.right) == 0:
        created.right = [EPSILON]
    return created

def remove_rules_with_epsilon(grammar: Grammar, transform_grammar=False) -> Grammar:
    # Copy if required
    if transform_grammar is False: grammar = copy(grammar)
    # Find nonterminals rewritable to epsilon
    rewritable = find_nonterminals_rewritable_to_epsilon(grammar)
    # Create list from rules
    rules = list(grammar.rules())
    index = 0
    # Iterate thought rules
    while index < len(rules):
        rule = rules[index]
        index += 1
        right = rule.right
        if right == [EPSILON]:
            if not grammar.start_isSet() or rule.fromSymbol != grammar.start_get():
                grammar.remove_rule(rule)
            # Continue IS executed, but due optimalization line is marked as missed.
            continue #pragma: no cover
        for rule_index in range(len(right)):
            symbol = right[rule_index]
            if symbol in rewritable:
                new_rule = _create_rule(rule, rule_index, rewritable)
                rules.append(new_rule)
                grammar.add_rule(new_rule)
    return grammar
